/*
 *  Copyright 2007 LAAS-CNRS
 *
 *  Authors: Florent Lamiraux, Luis Delgado
 */

#include <cerrno>
#include <iostream>

#include "KineoModel/kppFreeFlyerJointComponent.h"
#include "KineoModel/kppAnchorJointComponent.h"
#include "KineoModel/kppRotationJointComponent.h"
#include "KineoModel/kppTranslationJointComponent.h"

#include "MatrixAbstractLayer/MatrixAbstractLayer.h"
#include "hppModel/hppJoint.h"
#include "hppModel/hppDevice.h"
#include "hppModel/hppBody.h"
#include "kwsioConfig.h"

#if DEBUG==2
#undef NDEBUG
#define ODEBUG2(x) std::cout << "ChppDevice:" << x << std::endl
#define ODEBUG1(x) std::cerr << "ChppDevice:" << x << std::endl
#elif DEBUG==1
#undef NDEBUG
#define ODEBUG2(x)
#define ODEBUG1(x) std::cerr << "ChppDevice:" << x << std::endl
#else
#define ODEBUG2(x)
#define ODEBUG1(x)
#endif

// ==========================================================================

ChppDevice::ChppDevice(CjrlRobotDynamicsObjectFactory *inObjFactory) :
  CimplDynamicRobot(inObjFactory), CkppDeviceComponent()
{
}

// ==========================================================================

ChppDevice::ChppDevice() :
  CimplDynamicRobot(), CkppDeviceComponent()
{
}

// ==========================================================================

ChppDevice::~ChppDevice()
{
}

// ==========================================================================

ChppDeviceShPtr ChppDevice::create(std::string inName)
{
  CimplObjectFactory* objFactory = new CimplObjectFactory();
  ChppDevice *hppDevice = new ChppDevice(objFactory);
  ChppDeviceShPtr hppDeviceShPtr(hppDevice);

  if (hppDevice->init(hppDeviceShPtr, inName) != KD_OK) {
    hppDeviceShPtr.reset();
  }
  return hppDeviceShPtr;
}

// ==========================================================================

ChppDeviceShPtr ChppDevice::createCopy(const ChppDeviceShPtr& inDevice)
{
  ChppDevice* ptr = new ChppDevice(*inDevice);
  ChppDeviceShPtr deviceShPtr(ptr);

  if(KD_OK != ptr->init(deviceShPtr, inDevice))	{
    deviceShPtr.reset();
  }

  return deviceShPtr;
}

// ==========================================================================

CkwsDeviceShPtr ChppDevice::clone() const
{
  return ChppDevice::createCopy(attWeakPtr.lock());
}

// ==========================================================================

CkppComponentShPtr ChppDevice::cloneComponent() const
{
  return ChppDevice::createCopy(attWeakPtr.lock());
}

// ==========================================================================

bool ChppDevice::isComponentClonable() const
{
  return true;
}

// ==========================================================================

ktStatus ChppDevice::init(const ChppDeviceWkPtr& inDevWkPtr,
			  const std::string &inName)
{
  ktStatus success = CkppDeviceComponent::init(inDevWkPtr, inName);

  if(KD_OK == success) {
    attWeakPtr = inDevWkPtr;
  }
  return success;
}

// ==========================================================================

ktStatus ChppDevice::init(const ChppDeviceWkPtr& inWeakPtr,
			  const ChppDeviceShPtr& inDevice)
{
  ktStatus  success = CkppDeviceComponent::init(inWeakPtr, inDevice);

  if(KD_OK == success) {
    attWeakPtr = inWeakPtr;
  }

  return success;
}

// ==========================================================================

ktStatus
ChppDevice::axisAlignedBoundingBox (double& xMin, double& yMin, double& zMin,
				    double& xMax, double& yMax, double& zMax)
  const
{

  TBodyVector bodyVector;
  this->getBodyVector(bodyVector);
  unsigned int j=bodyVector.size();

  xMin=9999999;
  yMin=9999999;
  zMin=9999999;
  xMax=-9999999;
  yMax=-9999999;
  zMax=-9999999;

  for(unsigned int i=0; i<j; i++)
  {
    /*Dynamic cast*/
    CkwsKCDBodyShPtr a;
    a=KIT_DYNAMIC_PTR_CAST(CkwsKCDBody, bodyVector[i]);
    if(!a)
    {
      ODEBUG1(":axisAlignedBoundingBox: Error, the CkwsBody not of type CkwsKCDBody");
      return KD_ERROR;
    }
    computeBodyBoundingBox(a,xMin,yMin,zMin,xMax,yMax,zMax);
  }
  return KD_OK;
}

// ==========================================================================

ktStatus ChppDevice::ignoreDeviceForCollision (ChppDeviceShPtr inDevice )
{
  ktStatus status = KD_OK ;

  std::vector<CkcdObjectShPtr> inDeviceInnerObjectList;

  //
  // Get body vector of inDevice
  //
  CkwsDevice::TBodyVector inDeviceBodyVector;
  inDevice->getBodyVector(inDeviceBodyVector) ;

  //
  // For each body of inDevice,
  //
  for (unsigned int bodyId = 0 ;  bodyId < inDeviceBodyVector.size();
       bodyId++)  {
    CkwsKCDBodyShPtr kcdBody;
    if (kcdBody = KIT_DYNAMIC_PTR_CAST(CkwsKCDBody,
				       inDeviceBodyVector[bodyId])) {
      //
      // get the inner object list of the body
      //
      std::vector< CkcdObjectShPtr > kcdBodyInnerObjects =
	kcdBody->innerObjects() ;

      //
      // and put each inner object in inDeviceInnerObjectList
      //
      for (unsigned int objectId =0 ; objectId < kcdBodyInnerObjects.size() ;
	   objectId ++) {
	CkcdObjectShPtr kcdObject = kcdBodyInnerObjects[objectId] ;
	inDeviceInnerObjectList.push_back(kcdObject) ;
      }
    }
    else {
      ODEBUG1(":ignoreDeviceForCollision : body is not KCD body.");
      return KD_ERROR ;
    }
  }

  //
  // Set all inner objects of inDevice in the ignoredOuterObject list
  // of each body of this device
  //
  CkwsDevice::TBodyVector thisBodyVector;
  getBodyVector(thisBodyVector);

  //
  // For each body of this device
  //
  for (unsigned int bodyId = 0 ;  bodyId < thisBodyVector.size(); bodyId++)  {
    CkwsKCDBodyShPtr thisKcdBody;
    if (thisKcdBody = KIT_DYNAMIC_PTR_CAST(CkwsKCDBody,
					   thisBodyVector[bodyId])) {
      //
      // Get the list of ignored outer objects
      //
      std::vector<CkcdObjectShPtr> thisIgnoredObjectList =
	thisKcdBody->ignoredOuterObjects() ;

      //
      // Add to this list all the inner objects of inDevice
      //
      for( unsigned int objectId = 0 ;
	   objectId < inDeviceInnerObjectList.size() ; objectId++) {
	thisIgnoredObjectList.push_back(inDeviceInnerObjectList[objectId]) ;
      }

      //
      // Set this new list as ignored outer object list of this device body
      //
      thisKcdBody->ignoredOuterObjects(thisIgnoredObjectList) ;
    }
    else {
      ODEBUG1(":ignoreDeviceForCollision : body is not KCD body.");
      return KD_ERROR ;
    }
  }

  return status ;
}

// ==========================================================================

void ChppDevice::computeBodyBoundingBox(const CkwsKCDBodyShPtr& body,
					double& xMin, double& yMin,
					double& zMin, double& xMax,
					double& yMax, double& zMax) const
{
  std::vector<CkcdObjectShPtr> listObject = body->innerObjects();
  unsigned int j= listObject.size();
  for(unsigned int i=0; i<j; i++)
  {
    ckcdObjectBoundingBox(listObject[i],xMin,yMin,zMin,xMax,yMax,zMax);
  }
}

// ==========================================================================

void ChppDevice::ckcdObjectBoundingBox(const CkcdObjectShPtr& object,
				       double& xMin, double& yMin,
				       double& zMin, double& xMax,
				       double& yMax, double& zMax) const
{
  double x,y,z;

  object->boundingBox()->getHalfLengths(x, y, z) ;

  /*Matrices absolute et relative*/
  CkitMat4 matrixAbsolutePosition;
  CkitMat4 matrixRelativePosition;
  object->getAbsolutePosition(matrixAbsolutePosition);
  object->boundingBox()->getRelativePosition(matrixRelativePosition);

  /*Creer les points et change position points*/
  CkitMat4 matrixChangePosition =
    matrixAbsolutePosition*matrixRelativePosition;

  CkitPoint3 position[8];

  position[0]=matrixChangePosition*CkitPoint3( x, y, z);
  position[1]=matrixChangePosition*CkitPoint3( x, y,-z);
  position[2]=matrixChangePosition*CkitPoint3( x,-y, z);
  position[3]=matrixChangePosition*CkitPoint3(-x, y, z);
  position[4]=matrixChangePosition*CkitPoint3( x,-y,-z);
  position[5]=matrixChangePosition*CkitPoint3(-x,-y, z);
  position[6]=matrixChangePosition*CkitPoint3(-x, y,-z);
  position[7]=matrixChangePosition*CkitPoint3(-x,-y,-z);

  for(int i=0; i<8; i++)
  {
	if((position[i])[0]<xMin)
	{
	  xMin=(position[i])[0];
	}
	if((position[i])[1]<yMin)
	{
	  yMin=(position[i])[1];
	}
	if((position[i])[2]<zMin)
	{
	  zMin=(position[i])[2];
	}
	if((position[i])[0]>xMax)
	{
	  xMax=(position[i])[0];
	}
	if((position[i])[1]>yMax)
	{
	  yMax=(position[i])[1];
	}
	if((position[i])[2]>zMax)
	{
	  zMax=(position[i])[2];
	}
  }

}

// ==========================================================================

ktStatus ChppDevice::addObstacle(const CkcdObjectShPtr& inObject,
				 bool inDistanceComputation)
{
  // Get robot vector of bodies.
  CkwsDevice::TBodyVector bodyVector;
  getBodyVector(bodyVector);

  // Loop over bodies of robot.
  for (CkwsDevice::TBodyIterator bodyIter = bodyVector.begin();
       bodyIter < bodyVector.end(); bodyIter++) {
    // Try to cast body into CkwsKCDBody
    CkwsKCDBodyShPtr kcdBody;
    ChppBodyShPtr hppBody;
    if (kcdBody = KIT_DYNAMIC_PTR_CAST(CkwsKCDBody, *bodyIter)) {
      if(hppBody = KIT_DYNAMIC_PTR_CAST(ChppBody, kcdBody)) {
	hppBody->addOuterObject(inObject, inDistanceComputation);
      }
      else {
	std::vector< CkcdObjectShPtr > collisionList =
	  kcdBody->outerObjects();
	collisionList.push_back(inObject);
	kcdBody->outerObjects(collisionList);
      }
    }
    else {
      ODEBUG1(":addObstacle: body is not KCD body. Obstacle is not inserted.");
    }
  }
  return KD_OK;
}


// ==========================================================================

void ChppDevice::setRootJoint(ChppJoint* inJoint)
{
  /*
    Set joint as Kineo root joint
  */
  CkppDeviceComponent::rootJointComponent(inJoint->kppJoint());

  /*
    Set joint as robotDynamics root joint
  */
  CimplDynamicRobot::rootJoint(*(inJoint->jrlJoint()));
}

// ==========================================================================

ChppJoint* ChppDevice::getRootJoint()
{
  /*
    Get CkppJointComponent root joint
  */
  CkppJointComponentShPtr kppJointComponent = rootJointComponent();

  ChppJoint* joint = kppToHppJoint(kppJointComponent);
  return joint;
}

// ==========================================================================

void ChppDevice::registerJoint(ChppJoint* inHppJoint)
{
  /*
    Store hppJoint in maps
  */
  attKppToHppJointMap[inHppJoint->kppJoint()] = inHppJoint;
  attJrlToHppJointMap[inHppJoint->jrlJoint()] = inHppJoint;
}

// ==========================================================================

ChppJoint* ChppDevice::kppToHppJoint(CkppJointComponentShPtr inKppJoint)
{
  if (attKppToHppJointMap.count(inKppJoint) != 1) {
    ODEBUG1(":kppToHppJoint: no ChppJoint for this CkppJointComponent.");
    return NULL;
  }
  return attKppToHppJointMap[inKppJoint];
}

// ==========================================================================

ChppJoint* ChppDevice::jrlToHppJoint(CjrlJoint* inJrlJoint)
{
  if (attJrlToHppJointMap.count(inJrlJoint) != 1) {
    ODEBUG1(":jrlToHppJoint: no ChppJoint for this CjrlJoint.");
    return NULL;
  }
  return attJrlToHppJointMap[inJrlJoint];
}

// ==========================================================================


bool ChppDevice::kwsToJrlDynamicsDofValues(const std::vector<double>&
					   inKwsDofVector,
					   vectorN& outJrlDynamicsDofVector)
{
  /*
     Count the number of extra dofs of the CkppDeviceComponent
     since the first degrees of freedom of inKwsDofVector correspond to these extra-dofs.
  */
  unsigned int rankInCkwsConfig = countExtraDofs();
  std::vector< CkppJointComponentShPtr > kppJointVector;
  getJointComponentVector(kppJointVector);

  /*
    Output vectors should be of right size
  */
  KWS_PRECONDITION(outJrlDynamicsDofVector.size() == numberDof());

  /*
    Loop over CkppDeviceComponent joints
  */
  for (unsigned int iKppJoint=0; iKppJoint < kppJointVector.size();
       iKppJoint++) {
    CkppJointComponentShPtr kppJoint = kppJointVector[iKppJoint];

    /*
       Check if the joint is found in the associated map
    */
    if(attKppToHppJointMap[kppJoint] == NULL){
      // std::cout<<"joint "<<kppJoint->name()<<" "<<"... not found. "<<endl;
      continue;
    }

    /*
      Get associated CjrlJoint
    */
    CjrlJoint* jrlJoint = attKppToHppJointMap[kppJoint]->jrlJoint();
    unsigned int jrlRankInConfig = jrlJoint->rankInConfiguration();

    ODEBUG2(":kwsToJrlDynamicsDofValues: iKppJoint=" << kppJoint->name()
	    << " jrlRankInConfig=" << jrlRankInConfig);

    /*
      Check rank in configuration wrt  dimension.
    */
    unsigned int jointDim = kppJoint->countDofComponents();
    if (jrlRankInConfig+jointDim > inKwsDofVector.size()) {
      ODEBUG1
	(":kwsToJrlDynamicsDofValues: rank in configuration is more than configuration dimension(rank = "
	 << jrlRankInConfig << ", dof = " << jointDim << ").");
      ODEBUG1(":kwsToJrlDynamicsDofValues:   vectorN: "
	      << outJrlDynamicsDofVector);
      return false;
    }

    /*
      Cast joint into one of the possible types
    */
    if (CkppFreeFlyerJointComponentShPtr jointFF =
	KIT_DYNAMIC_PTR_CAST(CkppFreeFlyerJointComponent,
			     kppJoint)) {
      // Translations along x, y, z
      outJrlDynamicsDofVector[jrlRankInConfig] =
	inKwsDofVector[rankInCkwsConfig];
      outJrlDynamicsDofVector[jrlRankInConfig+1] =
	inKwsDofVector[rankInCkwsConfig+1];
      outJrlDynamicsDofVector[jrlRankInConfig+2] =
	inKwsDofVector[rankInCkwsConfig+2];
      double rx = inKwsDofVector[rankInCkwsConfig+3];
      double ry = inKwsDofVector[rankInCkwsConfig+4];
      double rz = inKwsDofVector[rankInCkwsConfig+5];
      // Convert KineoWorks rotation angles to roll, pitch, yaw
      double roll, pitch, yaw;
      YawPitchRollToRollPitchYaw(rx, ry, rz, roll, pitch, yaw);
      outJrlDynamicsDofVector[jrlRankInConfig+3] = roll;
      outJrlDynamicsDofVector[jrlRankInConfig+4] = pitch;
      outJrlDynamicsDofVector[jrlRankInConfig+5] = yaw;
      ODEBUG2("Joint value: " << outJrlDynamicsDofVector[jrlRankInConfig]
	      << ", "
	      << outJrlDynamicsDofVector[jrlRankInConfig+1] << ", "
	      << outJrlDynamicsDofVector[jrlRankInConfig+2] << ", "
	      << outJrlDynamicsDofVector[jrlRankInConfig+3] << ", "
	      << outJrlDynamicsDofVector[jrlRankInConfig+4] << ", "
	      << outJrlDynamicsDofVector[jrlRankInConfig+5]);
      rankInCkwsConfig += 6;
    }
    else if(CkppRotationJointComponentShPtr jointRot =
	    KIT_DYNAMIC_PTR_CAST(CkppRotationJointComponent,
				 kppJoint)) {
      outJrlDynamicsDofVector[jrlRankInConfig] =
	inKwsDofVector[rankInCkwsConfig];
      ODEBUG2("Joint value: " << outJrlDynamicsDofVector[jrlRankInConfig]);
      rankInCkwsConfig ++;
    }
    else if(CkppTranslationJointComponentShPtr jointTrans =
	    KIT_DYNAMIC_PTR_CAST(CkppTranslationJointComponent,
				 kppJoint)) {
      outJrlDynamicsDofVector[jrlRankInConfig] =
	inKwsDofVector[rankInCkwsConfig];
      ODEBUG2("Joint value: " << outJrlDynamicsDofVector[jrlRankInConfig]);
      rankInCkwsConfig ++;
    }
    else {
      ODEBUG1(":kwsToJrlDynamicsDofValues: unknown type of joint.");
      ODEBUG1(":kwsToJrlDynamicsDofValues:    vectorN: " <<
	      outJrlDynamicsDofVector);
      return false;
    }
  }

  ODEBUG2("hppSetCurrentConfig: outJrlDynamicsDofVector = "
	  << outJrlDynamicsDofVector);
  return true;
}


// ==========================================================================

bool
ChppDevice::jrlDynamicsToKwsDofValues(const vectorN& inJrlDynamicsDofVector,
				      std::vector<double>& outKwsDofVector)
{
  /*
     Count the number of extra dofs of the CkppDeviceComponent.
  */
  unsigned int rankInDofValues = countExtraDofs();

  std::vector< CkppJointComponentShPtr > kppJointVector;
  getJointComponentVector(kppJointVector);

  /*
    Output vectors should be of right size
  */
  KWS_PRECONDITION(outKwsDofVector.size() == countDofs());

  /*
    Loop over CkppDeviceComponent joints
  */
  for (unsigned int iKppJoint=0; iKppJoint < kppJointVector.size();
       iKppJoint++) {
    CkppJointComponentShPtr kppJoint = kppJointVector[iKppJoint];
    unsigned int jointDim = kppJoint->countDofComponents();
    /*
      Get associated CjrlJoint
    */
    CjrlJoint* jrlJoint = attKppToHppJointMap[kppJoint]->jrlJoint();
    unsigned int jrlRankInConfig = jrlJoint->rankInConfiguration();

    ODEBUG2(":jrlDynamicsToKwsDofValues: iKppJoint=" << kppJoint->name()
	    << " jrlRankInConfig=" << jrlRankInConfig);

    /*
      Check rank in configuration wrt  dimension.
    */
    if (jointDim > 0 && jrlRankInConfig+jointDim > inJrlDynamicsDofVector.size()) {
      ODEBUG1
	(":jrlDynamicsToKwsDofValues: rank in configuration is more than configuration dimension(rank = "
	 << jrlRankInConfig << ", dof = " << jointDim << ").");
      return false;
    }
    /*
      Cast joint into one of the possible types
    */
    if (CkppFreeFlyerJointComponentShPtr jointFF =
	KIT_DYNAMIC_PTR_CAST(CkppFreeFlyerJointComponent,
			     kppJoint)) {
      // Translations along x, y, z
      outKwsDofVector[rankInDofValues  ] =
	inJrlDynamicsDofVector[jrlRankInConfig  ];
      outKwsDofVector[rankInDofValues+1] =
	inJrlDynamicsDofVector[jrlRankInConfig+1];
      outKwsDofVector[rankInDofValues+2] =
	inJrlDynamicsDofVector[jrlRankInConfig+2];

      double roll = inJrlDynamicsDofVector[jrlRankInConfig+3];
      double pitch = inJrlDynamicsDofVector[jrlRankInConfig+4];
      double yaw = inJrlDynamicsDofVector[jrlRankInConfig+5];

      // Convert KineoWorks rotation angles to roll, pitch, yaw
      double rx, ry, rz;
      RollPitchYawToYawPitchRoll(roll, pitch, yaw, rx, ry, rz);

      outKwsDofVector[rankInDofValues+3] = rx;
      outKwsDofVector[rankInDofValues+4] = ry;
      outKwsDofVector[rankInDofValues+5] = rz;

      rankInDofValues+= 6;
    }
    else if(CkppRotationJointComponentShPtr jointRot =
	    KIT_DYNAMIC_PTR_CAST(CkppRotationJointComponent,
				 kppJoint)) {
      outKwsDofVector[rankInDofValues] =
	inJrlDynamicsDofVector[jrlRankInConfig];
      rankInDofValues++;
    }
    else if(CkppTranslationJointComponentShPtr jointTrans =
	    KIT_DYNAMIC_PTR_CAST(CkppTranslationJointComponent,
				 kppJoint)) {
      outKwsDofVector[rankInDofValues] =
	inJrlDynamicsDofVector[jrlRankInConfig];
      rankInDofValues++;
    }
    else if(CkppAnchorJointComponentShPtr jointAnchor =
	    KIT_DYNAMIC_PTR_CAST(CkppAnchorJointComponent, kppJoint)){
      // do nothing
    }
    else {
      ODEBUG1(":jrlDynamicsToKwsDofValues: unknown type of joint.");
      ODEBUG1(":jrlDynamicsToKwsDofValues:   vectorN: " <<
	      inJrlDynamicsDofVector);
      return false;
    }
  }
  return true;
}

// ==========================================================================

bool ChppDevice::hppSetCurrentConfig(const CkwsConfig& inConfig,
				     EwhichPart inUpdateWhat)
{
  bool updateGeom = (inUpdateWhat == GEOMETRIC || inUpdateWhat == BOTH);
  bool updateDynamic = (inUpdateWhat == DYNAMIC || inUpdateWhat == BOTH);

  if (updateGeom) {
    ODEBUG2("hppSetCurrentConfig: updating geometric part");
    ODEBUG2("hppSetCurrentConfig: inConfig = " << inConfig);


    if (CkppDeviceComponent::setCurrentConfig(inConfig) != KD_OK) {
      return false;
    }
  }
  if (updateDynamic) {
    /*
      Allocate a vector for CjrldynamicRobot configuration
      Dimension of vector is size of dynamic part of robot.
    */
    MAL_VECTOR_DIM(jrlConfig, double, numberDof());
    std::vector<double> kwsDofVector;
    inConfig.getDofValues(kwsDofVector);
    if (!kwsToJrlDynamicsDofValues(kwsDofVector, jrlConfig)) {
      return false;
    }

    if (!currentConfiguration(jrlConfig)) {
      return false;
    }
    if (!computeForwardKinematics()) {
	return false;
    }
  }
  return true;
}

// ==========================================================================

bool ChppDevice::hppSetCurrentConfig(const vectorN& inConfig,
				     EwhichPart inUpdateWhat)
{
  bool updateGeom = (inUpdateWhat == GEOMETRIC || inUpdateWhat == BOTH);
  bool updateDynamic = (inUpdateWhat == DYNAMIC || inUpdateWhat == BOTH);

  if (updateDynamic) {
    ODEBUG2("hppSetCurrentConfig: updating dynamic part");
    ODEBUG2("hppSetCurrentConfig: inConfig = " << inConfig);
    if (!currentConfiguration(inConfig)) {
      return false;
    }
    if (!computeForwardKinematics()) {
	return false;
    }
  }
  if (updateGeom) {
    std::vector<double> dofValues(countDofs());
    this->getCurrentDofValues(dofValues);
    if (!jrlDynamicsToKwsDofValues(inConfig, dofValues)) {
      return false;
    }

    return (CkppDeviceComponent::setCurrentDofValues(dofValues) == KD_OK);
  }
  return true;
}

void
ChppDevice::RollPitchYawToYawPitchRoll(const double& inRx, const double& inRy,
				       const double& inRz, double& outRx,
				       double& outRy, double& outRz)
{
  const double cRx=cos(inRx);
  const double sRx=sin(inRx);
  const double cRy=cos(inRy);
  const double sRy=sin(inRy);
  const double cRz=cos(inRz);
  const double sRz=sin(inRz);
  const double r00 = cRy*cRz;
  const double r01 = sRx*sRy*cRz - cRx*sRz;
  double r02 = cRx*sRy*cRz + sRx*sRz;
  const double r10 = cRy*sRz;
  const double r11 = sRx*sRy*sRz + cRx*cRz;
  const double r12 = cRx*sRy*sRz - sRx*cRz;
  const double r22 = cRx*cRy;

  // make sure that r02 is in [-1,1]
  if (r02 < -1.)
    r02 = -1.;
  else if (r02 > 1.)
    r02 = 1.;

  errno=0;
  outRy = asin(r02);
  // Check that asin call was successful.
  assert(errno == 0);

  double cosOutRy = cos(outRy);

  if(fabs(cosOutRy) > 1e-6) {
    double cosOutRx =  r22 / cosOutRy;
    double sinOutRx = -r12 / cosOutRy;

    outRx = atan2(sinOutRx, cosOutRx);

    double cosOutRz =  r00 / cosOutRy;
    double sinOutRz = -r01 / cosOutRy;
    outRz = atan2(sinOutRz, cosOutRz);
  }
  else {
    outRx = 0.;

    double cosOutRz = r11;
    double sinOutRz = r10;
    outRz = atan2(sinOutRz, cosOutRz);
  }
}

void
ChppDevice::YawPitchRollToRollPitchYaw(const double& inRx, const double& inRy,
				       const double& inRz, double& outRx,
				       double& outRy, double& outRz)
{
  const double cRx = cos(inRx);
  const double sRx = sin(inRx);
  const double cRy = cos(inRy);
  const double sRy = sin(inRy);
  const double cRz  =cos(inRz);
  const double sRz = sin(inRz);
  const double r00 = cRz * cRy;
  const double r01 = -sRz * cRy;
  const double r10 = cRz * sRy * sRx + sRz * cRx;
  const double r11 = cRz * cRx - sRz * sRy * sRx;
  double r20 = sRz * sRx - cRz * sRy * cRx;
  const double r21 = sRz * sRy * cRx + cRz * sRx;
  const double r22 = cRx * cRy;

  // make sure all values are in [-1,1]
  // as trigonometric functions would
  // fail when given values such as 1.00000001
  if (r20 < -1.)
    r20 = -1.;
  else if (r20 > 1.)
    r20 = 1.;

  errno=0;
  outRy = -asin(r20);
  // Check that asin call was successful.
  assert(errno == 0);

  double cosOutRy = cos(outRy);

  if (fabs(cosOutRy) > 1e-6) {
    double sinOutRx = r21 / cosOutRy;
    double cosOutRx = r22 / cosOutRy;
    outRx = atan2(sinOutRx, cosOutRx);

    double cosOutRz = r00 / cosOutRy;
    double sinOutRz = r10 / cosOutRy;
    outRz = atan2(sinOutRz, cosOutRz);
  } else {
    outRx = 0.;

    double cosOutRz = r11;
    double sinOutRz = -r01;
    outRz = atan2(sinOutRz, cosOutRz);
  }
}

// ==========================================================================

ChppJoint* ChppDevice::createFreeFlyer(std::string inName,
				       const CkitMat4& inInitialPosition)
{
  /*
    Create kppJointComponent
  */
  CkppJointComponentShPtr kppJoint = CkppFreeFlyerJointComponent::create(inName);
  if (kppJoint) {
    kppJoint->kwsJoint()->setCurrentPosition(inInitialPosition);
  } else {
    return NULL;
  }

  /*
    Convert homogeneous matrix to abstract matrix type matrix4d
  */
  matrix4d initialPos =
    ChppJoint::abstractMatrixFromCkitMat4(inInitialPosition);
  CimplObjectFactory objFactory;
  CjrlJoint* jrlJoint = objFactory.createJointFreeflyer(initialPos);
  if (!jrlJoint) {
    delete jrlJoint;
    return NULL;
  }

  /*
    Create ChppJoint
  */
  ChppJoint* hppJoint = new ChppJoint(kppJoint, jrlJoint, attWeakPtr);

  registerJoint(hppJoint);
  return hppJoint;
}

// ==========================================================================

ChppJoint* ChppDevice::createAnchor(std::string inName,
				    const CkitMat4& inInitialPosition)
{
  /*
    Create kppJointComponent
  */
  CkppJointComponentShPtr kppJoint = CkppAnchorJointComponent::create(inName);
  if (kppJoint) {
    kppJoint->kwsJoint()->setCurrentPosition(inInitialPosition);
  } else {
    return NULL;
  }

  /*
    Convert homogeneous matrix to abstract matrix type matrix4d
  */
  matrix4d initialPos =
    ChppJoint::abstractMatrixFromCkitMat4(inInitialPosition);
  CimplObjectFactory objFactory;
  CjrlJoint* jrlJoint = objFactory.createJointAnchor(initialPos);
  if (!jrlJoint) {
    delete jrlJoint;
    return NULL;
  }

  /*
    Create ChppJoint
  */
  ChppJoint* hppJoint = new ChppJoint(kppJoint, jrlJoint, attWeakPtr);

  registerJoint(hppJoint);
  return hppJoint;
}

// ==========================================================================

ChppJoint* ChppDevice::createRotation(std::string inName,
				      const CkitMat4& inInitialPosition)
{
  /*
    Create kppJointComponent
  */
  CkppJointComponentShPtr kppJoint = CkppRotationJointComponent::create(inName);
  if (kppJoint) {
    kppJoint->kwsJoint()->setCurrentPosition(inInitialPosition);
  } else {
    return NULL;
  }

  /*
    Convert homogeneous matrix to abstract matrix type matrix4d
  */
  matrix4d initialPos =
    ChppJoint::abstractMatrixFromCkitMat4(inInitialPosition);
  CimplObjectFactory objFactory;
  CjrlJoint* jrlJoint = objFactory.createJointRotation(initialPos);
  if (!jrlJoint) {
    delete jrlJoint;
    return NULL;
  }

  /*
    Create ChppJoint
  */
  ChppJoint* hppJoint = new ChppJoint(kppJoint, jrlJoint, attWeakPtr);

  registerJoint(hppJoint);
  return hppJoint;
}

// ==========================================================================

ChppJoint* ChppDevice::createTranslation(std::string inName,
					 const CkitMat4& inInitialPosition)
{
  /*
    Create kppJointComponent
  */
  CkppJointComponentShPtr kppJoint =
    CkppTranslationJointComponent::create(inName);
  if (kppJoint) {
    kppJoint->kwsJoint()->setCurrentPosition(inInitialPosition);
  } else {
    return NULL;
  }

  /*
    Convert homogeneous matrix to abstract matrix type matrix4d
  */
  matrix4d initialPos =
    ChppJoint::abstractMatrixFromCkitMat4(inInitialPosition);
  CimplObjectFactory objFactory;
  CjrlJoint* jrlJoint = objFactory.createJointTranslation(initialPos);
  if (!jrlJoint) {
    delete jrlJoint;
    return NULL;
  }

  /*
    Create ChppJoint
  */
  ChppJoint* hppJoint = new ChppJoint(kppJoint, jrlJoint, attWeakPtr);

  registerJoint(hppJoint);
  return hppJoint;
}

// ==========================================================================

/**
  \brief Write a device in a stream.
*/
std::ostream& operator<<(std::ostream& os, ChppDevice& inHppDevice)
{
  os << "Device: " << inHppDevice.name() << std::endl;
  os << std::endl;
  os << "  Current configuration: " << inHppDevice.currentConfiguration()
     << std::endl;
  os << std::endl;
  os << std::endl;
  os << "  Writing kinematic chain" << std::endl;

  //
  // Go through joints and output each joint
  //
  ChppJoint* hppJoint = inHppDevice.getRootJoint();

  if (hppJoint) {
    os << *hppJoint << std::endl;
  }
  // Get position of center of mass
  MAL_S3_VECTOR(com, double);
  com = inHppDevice.positionCenterOfMass();

  //debug
  std::cout<<"total mass "<<inHppDevice.mass() <<", COM: "
      << MAL_S3_VECTOR_ACCESS(com, 0)
      <<", "<< MAL_S3_VECTOR_ACCESS(com, 1)
      <<", "<< MAL_S3_VECTOR_ACCESS(com, 2)
      <<std::endl;
  return os;
}
